/*-
 * Copyright (c) 2011-2012 George V. Neville-Neil,
 *                         Steven Kreuzer, 
 *                         Martin Burnicki, 
 *                         Jan Breuer,
 *                         Gael Mace, 
 *                         Alexandre Van Kempen,
 *                         Inaqui Delgado,
 *                         Rick Ratzel,
 *                         National Instruments.
 * Copyright (c) 2009-2010 George V. Neville-Neil, 
 *                         Steven Kreuzer, 
 *                         Martin Burnicki, 
 *                         Jan Breuer,
 *                         Gael Mace, 
 *                         Alexandre Van Kempen
 *
 * Copyright (c) 2005-2008 Kendall Correll, Aidan Williams
 *
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <netinet/ip.h>
#include <arpa/inet.h>

#include <exanic/fifo_tx.h>
#include <exanic/fifo_rx.h>
#include <exanic/exanic.h>
#include <exanic/port.h>
#include <exanic/config.h>
#include <exanic/time.h>
#include <exanic/register.h>
#include "../ptpd.h"
#include "exanic.h"
#include "net.h"

static uint8_t ifaceMAC[6];
static const uint8_t multicastMAC[] = {0x01, 0x00, 0x5E, 0x00, 0x01, 0x81 };
static NetPath *pLocalNetPath;

static void processFrame(struct etherFrameType *frame, uint32_t *timestamp );

void calcIPHdrChecksum( uint8_t *buf, uint16_t *checksum, int length )
{
    int i;
    uint32_t sum;
    uint16_t csum_16;

    sum=0;
    for( i=0; i<=length; i+=2 )
    {
        if( i==10 )
            continue;       // skip checksum field
        sum += (buf[i] << 8) + buf[i+1];
    }
    csum_16 = sum & 0xFFFF;
    csum_16 += sum >> 16;
    csum_16 = ~csum_16;
    *checksum = csum_16 >> 8;
    *checksum += (csum_16 & 0xFF) << 8;
}

int checkIPHdrChecksum( uint8_t *pkt )
{
    uint32_t sum=0, i;

    for(i=0; i<20; i+=2 )
        sum += (pkt[i]<<8) + pkt[i+1];

    sum = (sum & 0xFFFF) + (sum >> 16);
    if( sum == 0xFFFF )
        return 0;   // all OK
    return -1;
}
/**
  * @brief  Initialize network queue
  * @param  pQ the queue to be initialized
  * @retval None
  */
static void netQInit(BufQueue *pQ)
{
    pQ->get = 0;
    pQ->put = 0;
    pQ->count = 0;
}

/**
  * @brief  Put data to the network queue
  * @param  pQ the queue to be used
  * @param  pbuf the packet to be put to the queue
  * @retval Boolean success
  */
static Boolean netQPut(BufQueue *pQ, void *pbuf)
{
    if (pQ->count >= PBUF_QUEUE_SIZE)
        return FALSE;

    pQ->pbuf[pQ->put] = pbuf;

    pQ->put = (pQ->put + 1) % PBUF_QUEUE_SIZE;

    pQ->count++;

    return TRUE;
}

/**
  * @brief  Get data from the network queue
  * @param  pQ the queue to be used
  * @retval void* pointer to pbuf or NULL
  */
static void *netQGet(BufQueue *pQ)
{
    void *pbuf;

    if (!pQ->count)
        return NULL;

    pbuf = pQ->pbuf[pQ->get];

    pQ->get = (pQ->get + 1) % PBUF_QUEUE_SIZE;

    pQ->count--;

    return pbuf;
}

/**
  * @brief  Free all pbufs in the queue
  * @param  pQ the queue to be used
  * @retval None
  */
static void netQEmpty(BufQueue * pQ)
{

    struct pbuf * p;
    int cnt = pQ->count;

    for (;cnt > 0; cnt--)
    {
        p = (struct pbuf*)netQGet(pQ);

        if (p) ;//pbuf_free(p);
    }
}

/**
  * @brief  Check if something is in the queue
  * @param  pQ the queue to be used
  * @retval Boolean TRUE if success
  */
static Boolean netQCheck(BufQueue *pQ)
{
    if (!pQ->count)
        return FALSE;

    return TRUE;
}

Boolean
chooseMcastGroup(RunTimeOpts * rtOpts, struct in_addr *netAddr)
{

    char *addrStr;

#ifdef PTP_EXPERIMENTAL
    switch(rtOpts->mcast_group_Number){
    case 0:
        addrStr = DEFAULT_PTP_DOMAIN_ADDRESS;
        break;

    case 1:
        addrStr = ALTERNATE_PTP_DOMAIN1_ADDRESS;
        break;
    case 2:
        addrStr = ALTERNATE_PTP_DOMAIN2_ADDRESS;
        break;
    case 3:
        addrStr = ALTERNATE_PTP_DOMAIN3_ADDRESS;
        break;

    default:
        ERROR("Unk group %d\n", rtOpts->mcast_group_Number);
        exit(3);
        break;
    }
#else
    addrStr = DEFAULT_PTP_DOMAIN_ADDRESS;
#endif

    if (!inet_aton(addrStr, netAddr)) {
        ERROR("failed to encode multicast address: %s\n", addrStr);
        return FALSE;
    }
    return TRUE;
}

/**
 * Init the multcast for specific IPv4 address
 *
 * @param netPath
 * @param multicastAddr
 *
 * @return TRUE if successful
 */
static Boolean
netInitMulticastIPv4(NetPath * netPath, Integer32 multicastAddr)
{
    struct ip_mreq imr;

    if( netPath->ifaceEndpoint->exanic != NULL )
    {

    }
    else
    {
        /* multicast send only on specified interface */
        imr.imr_multiaddr.s_addr = multicastAddr;
        imr.imr_interface.s_addr = netPath->interfaceAddr.s_addr;
        if (setsockopt(netPath->eventSock, IPPROTO_IP, IP_MULTICAST_IF,
                   &imr.imr_interface.s_addr, sizeof(struct in_addr)) < 0
            || setsockopt(netPath->generalSock, IPPROTO_IP, IP_MULTICAST_IF,
                  &imr.imr_interface.s_addr, sizeof(struct in_addr))
            < 0) {
            PERROR("failed to enable multi-cast on the interface");
            return FALSE;
        }
        /* join multicast group (for receiving) on specified interface */
        if (setsockopt(netPath->eventSock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                   &imr, sizeof(struct ip_mreq)) < 0
            || setsockopt(netPath->generalSock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                  &imr, sizeof(struct ip_mreq)) < 0) {
            PERROR("failed to join the multi-cast group");
            return FALSE;
        }
    }
    return TRUE;
}

/**
 * Init the multcast (both General and Peer)
 *
 * @param netPath
 * @param rtOpts
 *
 * @return TRUE if successful
 */
Boolean
netInitMulticast(NetPath * netPath,  RunTimeOpts * rtOpts)
{
    struct in_addr netAddr;
    char addrStr[NET_ADDRESS_LENGTH];

    /* Init General multicast IP address */
    if(!chooseMcastGroup(rtOpts, &netAddr)){
        return FALSE;
    }

    netPath->multicastAddr = netAddr.s_addr;
    if(!netInitMulticastIPv4(netPath, netPath->multicastAddr)) {
        return FALSE;
    }
    /* End of General multicast Ip address init */


    /* Init Peer multicast IP address */
    memcpy(addrStr, PEER_PTP_DOMAIN_ADDRESS, NET_ADDRESS_LENGTH);

    if (!inet_aton(addrStr, &netAddr)) {
        ERROR("failed to encode multi-cast address: %s\n", addrStr);
        return FALSE;
    }
    netPath->peerMulticastAddr = netAddr.s_addr;
    if(!netInitMulticastIPv4(netPath, netPath->peerMulticastAddr)) {
        return FALSE;
    }
    /* End of Peer multicast Ip address init */

    return TRUE;
}

/**
 * Initialize timestamping of packets
 *
 * @param netPath
 *
 * @return TRUE if successful
 */
Boolean
netInitTimestamping(NetPath * netPath)
{
    int val = 1;
    Boolean result = TRUE;

#if defined(SO_TIMESTAMPNS) /* Linux, Apple */
    DBG("netInitTimestamping: trying to use SO_TIMESTAMPNS\n");

    if (setsockopt(netPath->eventSock, SOL_SOCKET, SO_TIMESTAMPNS, &val, sizeof(int)) < 0
        || setsockopt(netPath->generalSock, SOL_SOCKET, SO_TIMESTAMPNS, &val, sizeof(int)) < 0) {
        PERROR("netInitTimestamping: failed to enable SO_TIMESTAMPNS");
        result = FALSE;
    }
#elif defined(SO_BINTIME) /* FreeBSD */
    DBG("netInitTimestamping: trying to use SO_BINTIME\n");

    if (setsockopt(netPath->eventSock, SOL_SOCKET, SO_BINTIME, &val, sizeof(int)) < 0
        || setsockopt(netPath->generalSock, SOL_SOCKET, SO_BINTIME, &val, sizeof(int)) < 0) {
        PERROR("netInitTimestamping: failed to enable SO_BINTIME");
        result = FALSE;
    }
#else
    result = FALSE;
#endif

/* fallback method */
#if defined(SO_TIMESTAMP) /* Linux, Apple, FreeBSD */
    if (!result) {
        DBG("netInitTimestamping: trying to use SO_TIMESTAMP\n");

        if (setsockopt(netPath->eventSock, SOL_SOCKET, SO_TIMESTAMP, &val, sizeof(int)) < 0
            || setsockopt(netPath->generalSock, SOL_SOCKET, SO_TIMESTAMP, &val, sizeof(int)) < 0) {
            PERROR("netInitTimestamping: failed to enable SO_TIMESTAMP");
            result = FALSE;
        }
        result = TRUE;
    }
#endif

    return result;
}

static Boolean
netShutdownMulticastIPv4(NetPath * netPath, Integer32 multicastAddr)
{
    struct ip_mreq imr;

    /* Close General Multicast */
    imr.imr_multiaddr.s_addr = multicastAddr;
    imr.imr_interface.s_addr = netPath->interfaceAddr.s_addr;

    setsockopt(netPath->eventSock, IPPROTO_IP, IP_DROP_MEMBERSHIP,
           &imr, sizeof(struct ip_mreq));
    setsockopt(netPath->generalSock, IPPROTO_IP, IP_DROP_MEMBERSHIP,
           &imr, sizeof(struct ip_mreq));

    return TRUE;
}

/**
 * shutdown the multicast (both General and Peer)
 *
 * @param netPath
 *
 * @return TRUE if successful
 */
static Boolean
netShutdownMulticast(NetPath * netPath)
{
    /* Close General Multicast */
    netShutdownMulticastIPv4(netPath, netPath->multicastAddr);
    netPath->multicastAddr = 0;

    /* Close Peer Multicast */
    netShutdownMulticastIPv4(netPath, netPath->peerMulticastAddr);
    netPath->peerMulticastAddr = 0;

    return TRUE;
}

/**
  * @brief  shut down the UDP and network stuff
  * @retval TRUE if success
  */
Boolean netShutdown(NetPath *netPath)
{

    /* Return a success code. */
    if( netPath->ifaceEndpoint != NULL )
        endpoint_close(netPath->ifaceEndpoint);
    return TRUE;
}

/**
  * @brief  Find interface to be used
  * @param  ifaceName name of the required interface (will be filled with appropriate interface name if not set)
  * @param  uuid will be filled with MAC address of the interface
  * @param  netPath network object
  * @retval Integer32 IPv4 address of the interface
  */
static in_addr_t findIface(const Octet *ifaceName, Octet *uuid, NetPath *netPath)
{
    static struct endpoint e;
    struct ifreq device;
    struct in_addr addr;
    exanic_if_addr_t addr_t;

    if( netPath->ifaceEndpoint != NULL )
        endpoint_close(netPath->ifaceEndpoint);

    if( endpoint_open(&e, ifaceName, 1, 1) != 0 )
    {
        ERROR("findIface(): Failed to open endpoint %s\n", ifaceName );
        return 0;
    }
    netPath->ifaceEndpoint = &e;
    if( netPath->ifaceEndpoint->exanic != NULL )
    {
        if( exanic_get_mac_addr(netPath->ifaceEndpoint->exanic,netPath->ifaceEndpoint->rx->port_number, ifaceMAC ) == -1 )
        {
            ERROR("findIface(): Failed to get MAC address\n");
            return 0;
        }
        memcpy( uuid, ifaceMAC, PTP_UUID_LENGTH);
        if( exanic_get_interface_addr(netPath->ifaceEndpoint->exanic,netPath->ifaceEndpoint->rx->port_number, &addr_t ) == -1 )
        {
            ERROR("findIface(): Failed to get IP address\n");
            return 0;
        }
        return addr_t.address;
    }
    else
    {
    	/* open sockets */
    	if ((netPath->eventSock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0
    	    || (netPath->generalSock = socket(PF_INET, SOCK_DGRAM,
    					      IPPROTO_UDP)) < 0) {
    		PERROR("failed to initalize sockets");
    		return FALSE;
    	}

        memcpy(device.ifr_name, ifaceName, IFACE_NAME_LENGTH);

        if (ioctl(netPath->eventSock, SIOCGIFHWADDR, &device) < 0)
            DBGV("failed to get hardware address\n");
        else
            memcpy(uuid, device.ifr_hwaddr.sa_data, PTP_UUID_LENGTH);

        if (ioctl(netPath->eventSock, SIOCGIFADDR, &device) < 0) {
    		PERROR("failed to get ip address");
    		return 0;
    	}
        addr = (struct in_addr)(((struct sockaddr_in *)&device.ifr_addr)->sin_addr);
        return addr.s_addr;
    }
}

/**
  * @brief  Start all of the UDP stuff
  * @param  netPath network object
  * @param  ptpClock PTP clock object
  * @retval Boolean success
  */
Boolean netInit(NetPath *netPath, RunTimeOpts* rtOpts, PtpClock *ptpClock)
{
    in_addr_t interfaceAddr;

    struct in_addr netAddr;
    struct sockaddr_in addr;
    int temp;

    DBG("netInit\n");

    /* open the network interface */
    interfaceAddr = findIface(rtOpts->ifaceName, ptpClock->port_uuid_field, netPath);

    if (!interfaceAddr)
    {
        ERROR("netInit(): Failed to open interface %s\n", rtOpts->ifaceName);
        return FALSE;
    }

    netPath->interfaceAddr.s_addr = interfaceAddr;
    DBG("Local IP address used : %s \n", inet_ntoa(netPath->interfaceAddr));

    pLocalNetPath = netPath;

    /* Initialize the buffer queues. */
    netQInit(&netPath->eventQ);

    netQInit(&netPath->generalQ);

    /* Configure network (broadcast/unicast) addresses. */
    netPath->unicastAddr = 0; /* disable unicast */

    /* init UDP Multicast on both Default and Pear addresses */
    if (!netInitMulticast(netPath, rtOpts)) {
        return FALSE;
    }

    if( netPath->ifaceEndpoint->exanic == NULL )
    {
        temp = 1;           /* allow address reuse */
        if (setsockopt(netPath->eventSock, SOL_SOCKET, SO_REUSEADDR,
                   &temp, sizeof(int)) < 0
            || setsockopt(netPath->generalSock, SOL_SOCKET, SO_REUSEADDR,
                  &temp, sizeof(int)) < 0) {
            DBG("failed to set socket reuse\n");
        }
        /* bind sockets */
        /*
         * need INADDR_ANY to allow receipt of multi-cast and uni-cast
         * messages
         */
        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = htonl(INADDR_ANY);
        addr.sin_port = htons(PTP_EVENT_PORT);
        if (bind(netPath->eventSock, (struct sockaddr *)&addr,
             sizeof(struct sockaddr_in)) < 0) {
            PERROR("failed to bind event socket");
            return FALSE;
        }
        addr.sin_port = htons(PTP_GENERAL_PORT);
        if (bind(netPath->generalSock, (struct sockaddr *)&addr,
             sizeof(struct sockaddr_in)) < 0) {
            PERROR("failed to bind general socket");
            return FALSE;
        }

        /*
         * The following code makes sure that the data is only received on the specified interface.
         * Without this option, it's possible to receive PTP from another interface, and confuse the protocol.
         * Calling bind() with the IP address of the device instead of INADDR_ANY does not work.
         *
         * More info:
         *   http://developerweb.net/viewtopic.php?id=6471
         *   http://stackoverflow.com/questions/1207746/problems-with-so-bindtodevice-linux-socket-option
         */
    #ifdef PTP_EXPERIMENTAL
        if ( !rtOpts->do_hybrid_mode )
    #endif
        if (setsockopt(netPath->eventSock, SOL_SOCKET, SO_BINDTODEVICE,
                rtOpts->ifaceName, strlen(rtOpts->ifaceName)) < 0
            || setsockopt(netPath->generalSock, SOL_SOCKET, SO_BINDTODEVICE,
                rtOpts->ifaceName, strlen(rtOpts->ifaceName)) < 0){
            PERROR("failed to call SO_BINDTODEVICE on the interface");
            return FALSE;
        }

        /* send a uni-cast address if specified (useful for testing) */
        if (rtOpts->unicastAddress[0]) {
            /* Attempt a DNS lookup first. */
            struct hostent *host;
            host = gethostbyname2(rtOpts->unicastAddress, AF_INET);
                    if (host != NULL) {
                if (host->h_length != 4) {
                    PERROR("unicast host resolved to non ipv4"
                           "address");
                    return FALSE;
                }
                netPath->unicastAddr =
                    *(uint32_t *)host->h_addr_list[0];
            } else {
                /* Maybe it's a dotted quad. */
                if (!inet_aton(rtOpts->unicastAddress, &netAddr)) {
                    ERROR("failed to encode uni-cast address: %s\n",
                          rtOpts->unicastAddress);
                    return FALSE;
                    netPath->unicastAddr = netAddr.s_addr;
                }
                    }
        } else {
                    netPath->unicastAddr = 0;
        }

        /* set socket time-to-live to 1 */

        if (setsockopt(netPath->eventSock, IPPROTO_IP, IP_MULTICAST_TTL,
                   &rtOpts->ttl, sizeof(int)) < 0
            || setsockopt(netPath->generalSock, IPPROTO_IP, IP_MULTICAST_TTL,
                  &rtOpts->ttl, sizeof(int)) < 0) {
            PERROR("failed to set the multi-cast time-to-live");
            return FALSE;
        }

        /* enable loopback */
        temp = 1;

        DBG("Going to set IP_MULTICAST_LOOP with %d \n", temp);

        if (setsockopt(netPath->eventSock, IPPROTO_IP, IP_MULTICAST_LOOP,
                   &temp, sizeof(int)) < 0
            || setsockopt(netPath->generalSock, IPPROTO_IP, IP_MULTICAST_LOOP,
                  &temp, sizeof(int)) < 0) {
            PERROR("failed to enable multi-cast loopback");
            return FALSE;
        }

        /* make timestamps available through recvmsg() */
        if (!netInitTimestamping(netPath)) {
            ERROR("failed to enable receive time stamps");
            return FALSE;
        }
    }


    /* Return a success code. */
    return TRUE;
}

/**
  * @brief  Wait for a packet to come in on either port.  For now, there is no wait.
  * Simply check to see if a packet is available on either port and return 1,
  * otherwise return 0.
  * @param  netPath network object
  * @param  timeout not used
  * @retval Integer32 number > 0 if there are some data
  */
Integer32 netSelect(TimeInternal* notUsed, NetPath *netPath)
{
    /* Check the packet queues.  If there is data, return TRUE. */
    int ret, nfds;
    fd_set readfds;

    // If we're bound to a exanic
    if( netPath->ifaceEndpoint->exanic != NULL )
    {
        if (netQCheck(&netPath->eventQ) || netQCheck(&netPath->generalQ))
            return 1;
        return 0;
    }

    // else process a standard NIC interface
    FD_ZERO(&readfds);
    FD_SET(netPath->eventSock, &readfds);
    FD_SET(netPath->generalSock, &readfds);

    if (netPath->eventSock > netPath->generalSock)
        nfds = netPath->eventSock;
    else
        nfds = netPath->generalSock;

    ret = select(nfds + 1, &readfds, 0, 0, NULL) > 0;

    if (ret < 0) {
        if (errno == EAGAIN || errno == EINTR)
            return 0;
    }
    return ret;
}

/**
  * @brief  Delete all waiting packets in Event queue
  * @param  netPath network object
  * @retval None
  */
void netEmptyEventQ(NetPath *netPath)
{
    netQEmpty(&netPath->eventQ);
}

static ssize_t netRecv(Octet *buf, TimeInternal *time, BufQueue * msgQueue)
{
    struct ptpPacketType *pkt;
    int len;

    pkt = (struct ptpPacketType*)netQGet(msgQueue);
    if (!pkt || !buf)
        return 0;

    len = pkt->len;
    if( len > PACKET_SIZE )
    {
        ERROR("netRecv: received truncated message\n");
        return 0;
    }

    if(time)
    {
        time->seconds = pkt->ts_secs;
        time->nanoseconds = pkt->ts_nanos;
    }
    memcpy( buf, pkt->buf, pkt->len );
    free(pkt->buf);
    free(pkt);
    return len;
}

ssize_t netRecvEvent(Octet *buf, TimeInternal *time, NetPath *netPath)
{
    if( netPath->ifaceEndpoint->exanic != NULL )
        return netRecv(buf, time, &netPath->eventQ);
    else
    {
        // process standard NIC interface
        ssize_t ret;
        struct msghdr msg;
        struct iovec vec[1];
        struct sockaddr_in from_addr;

        union {
            struct cmsghdr cm;
            char    control[CMSG_SPACE(sizeof(struct timeval))];
        }     cmsg_un;

        struct cmsghdr *cmsg;

    #if defined(SO_TIMESTAMPNS)
        struct timespec * ts;
    #elif defined(SO_BINTIME)
        struct bintime * bt;
        struct timespec ts;
    #endif

    #if defined(SO_TIMESTAMP)
        struct timeval * tv;
    #endif
        Boolean timestampValid = FALSE;


        vec[0].iov_base = buf;
        vec[0].iov_len = PACKET_SIZE;

        memset(&msg, 0, sizeof(msg));
        memset(&from_addr, 0, sizeof(from_addr));
        memset(buf, 0, PACKET_SIZE);
        memset(&cmsg_un, 0, sizeof(cmsg_un));

        msg.msg_name = (caddr_t)&from_addr;
        msg.msg_namelen = sizeof(from_addr);
        msg.msg_iov = vec;
        msg.msg_iovlen = 1;
        msg.msg_control = cmsg_un.control;
        msg.msg_controllen = sizeof(cmsg_un.control);
        msg.msg_flags = 0;

        ret = recvmsg(netPath->eventSock, &msg, MSG_DONTWAIT);
        if (ret <= 0) {
            if (errno == EAGAIN || errno == EINTR)
                return 0;

            return ret;
        }
        if (msg.msg_flags & MSG_TRUNC) {
            ERROR("received truncated message\n");
            return 0;
        }
        /* get time stamp of packet */
        if (!time) {
            ERROR("null receive time stamp argument\n");
            return 0;
        }
        if (msg.msg_flags & MSG_CTRUNC) {
            ERROR("received truncated ancillary data\n");
            return 0;
        }
    #ifdef PTP_EXPERIMENTAL
        netPath->lastRecvAddr = from_addr.sin_addr.s_addr;
    #endif

        if (msg.msg_controllen <= 0) {
            ERROR("received short ancillary data (%ld/%ld)\n",
                (long)msg.msg_controllen, (long)sizeof(cmsg_un.control));

            return 0;
        }

        for (cmsg = CMSG_FIRSTHDR(&msg); cmsg != NULL;
             cmsg = CMSG_NXTHDR(&msg, cmsg)) {
            if (cmsg->cmsg_level == SOL_SOCKET) {
    #if defined(SO_TIMESTAMPNS)
                if(cmsg->cmsg_type == SCM_TIMESTAMPNS) {
                    ts = (struct timespec *)CMSG_DATA(cmsg);
                    time->seconds = ts->tv_sec;
                    time->nanoseconds = ts->tv_nsec;
                    timestampValid = TRUE;
                    DBGV("kernel NANO recv time stamp %us %dns\n",
                         time->seconds, time->nanoseconds);
                    break;
                }
    #elif defined(SO_BINTIME)
                if(cmsg->cmsg_type == SCM_BINTIME) {
                    bt = (struct bintime *)CMSG_DATA(cmsg);
                    bintime2timespec(bt, &ts);
                    time->seconds = ts.tv_sec;
                    time->nanoseconds = ts.tv_nsec;
                    timestampValid = TRUE;
                    DBGV("kernel NANO recv time stamp %us %dns\n",
                         time->seconds, time->nanoseconds);
                    break;
                }
    #endif

    #if defined(SO_TIMESTAMP)
                if(cmsg->cmsg_type == SCM_TIMESTAMP) {
                    tv = (struct timeval *)CMSG_DATA(cmsg);
                    time->seconds = tv->tv_sec;
                    time->nanoseconds = tv->tv_usec * 1000;
                    timestampValid = TRUE;
                    DBGV("kernel MICRO recv time stamp %us %dns\n",
                         time->seconds, time->nanoseconds);
                }
    #endif
            }
        }

        if (!timestampValid) {
            /*
             * do not try to get by with recording the time here, better
             * to fail because the time recorded could be well after the
             * message receive, which would put a big spike in the
             * offset signal sent to the clock servo
             */
            DBG("netRecvEvent: no receive time stamp\n");
            return 0;
        }

        return ret;
    }
}

ssize_t netRecvGeneral(Octet *buf, TimeInternal *time, NetPath *netPath )
{
    if( netPath->ifaceEndpoint->exanic != NULL )
        return netRecv(buf, time, &netPath->generalQ);
    else
    {
        ssize_t ret;
        struct msghdr msg;
        struct iovec vec[1];
        struct sockaddr_in from_addr;

        union {
            struct cmsghdr cm;
            char    control[CMSG_SPACE(sizeof(struct timeval))];
        }     cmsg_un;

        struct cmsghdr *cmsg;

    #if defined(SO_TIMESTAMPNS)
        struct timespec * ts;
    #elif defined(SO_BINTIME)
        struct bintime * bt;
        struct timespec ts;
    #endif

    #if defined(SO_TIMESTAMP)
        struct timeval * tv;
    #endif
        Boolean timestampValid = FALSE;

        vec[0].iov_base = buf;
        vec[0].iov_len = PACKET_SIZE;

        memset(&msg, 0, sizeof(msg));
        memset(&from_addr, 0, sizeof(from_addr));
        memset(buf, 0, PACKET_SIZE);
        memset(&cmsg_un, 0, sizeof(cmsg_un));

        msg.msg_name = (caddr_t)&from_addr;
        msg.msg_namelen = sizeof(from_addr);
        msg.msg_iov = vec;
        msg.msg_iovlen = 1;
        msg.msg_control = cmsg_un.control;
        msg.msg_controllen = sizeof(cmsg_un.control);
        msg.msg_flags = 0;

        ret = recvmsg(netPath->generalSock, &msg, MSG_DONTWAIT);
        if (ret <= 0) {
            if (errno == EAGAIN || errno == EINTR)
                return 0;

            return ret;
        }
        if (msg.msg_flags & MSG_TRUNC) {
            ERROR("received truncated message\n");
            return 0;
        }
        /* get time stamp of packet */
        if (!time) {
            ERROR("null receive time stamp argument\n");
            return 0;
        }
        if (msg.msg_flags & MSG_CTRUNC) {
            ERROR("received truncated ancillary data\n");
            return 0;
        }

    #ifdef PTP_EXPERIMENTAL
        netPath->lastRecvAddr = from_addr.sin_addr.s_addr;
    #endif



        if (msg.msg_controllen <= 0) {
            ERROR("received short ancillary data (%ld/%ld)\n",
                  (long)msg.msg_controllen, (long)sizeof(cmsg_un.control));

            return 0;
        }

        for (cmsg = CMSG_FIRSTHDR(&msg); cmsg != NULL;
             cmsg = CMSG_NXTHDR(&msg, cmsg)) {
            if (cmsg->cmsg_level == SOL_SOCKET) {
    #if defined(SO_TIMESTAMPNS)
                if(cmsg->cmsg_type == SCM_TIMESTAMPNS) {
                    ts = (struct timespec *)CMSG_DATA(cmsg);
                    time->seconds = ts->tv_sec;
                    time->nanoseconds = ts->tv_nsec;
                    timestampValid = TRUE;
                    DBGV("kernel NANO recv time stamp %us %dns\n",
                         time->seconds, time->nanoseconds);
                    break;
                }
    #elif defined(SO_BINTIME)
                if(cmsg->cmsg_type == SCM_BINTIME) {
                    bt = (struct bintime *)CMSG_DATA(cmsg);
                    bintime2timespec(bt, &ts);
                    time->seconds = ts.tv_sec;
                    time->nanoseconds = ts.tv_nsec;
                    timestampValid = TRUE;
                    DBGV("kernel NANO recv time stamp %us %dns\n",
                         time->seconds, time->nanoseconds);
                    break;
                }
    #endif

    #if defined(SO_TIMESTAMP)
                if(cmsg->cmsg_type == SCM_TIMESTAMP) {
                    tv = (struct timeval *)CMSG_DATA(cmsg);
                    time->seconds = tv->tv_sec;
                    time->nanoseconds = tv->tv_usec * 1000;
                    timestampValid = TRUE;
                    DBGV("kernel MICRO recv time stamp %us %dns\n",
                         time->seconds, time->nanoseconds);
                }
    #endif
            }
        }

        if (!timestampValid) {
            /*
             * do not try to get by with recording the time here, better
             * to fail because the time recorded could be well after the
             * message receive, which would put a big spike in the
             * offset signal sent to the clock servo
             */
            DBG("netRecvGeneral: no receive time stamp\n");
            return 0;
        }

        return ret;
    }
}

static ssize_t netSend(const Octet *buf, UInteger16 length, const Integer32 * addr, NetPath *netPath, uint16_t port )
{
    uint16_t tmp;
    exanic_if_addr_t if_addr;
    int frameLen, ret;
    struct etherFrameType frame;
    uint32_t ts_tx_old, ts_tx;

    memset(&frame, 0, sizeof(frame));
    memcpy(frame.dstmac, multicastMAC, 6);
    memcpy(frame.srcmac, ifaceMAC, 6 );
    frame.etherType = htons(0x800);    // IP
    frame.ip.byte1 = 0x45;                 // IPv4
    frame.ip.length = htons(length + 20 + 8);
    frame.ip.flagsFrag = 0x40;
    frame.ip.ttl = 1;
    frame.ip.protocol = 0x11;              // UDP
    if (exanic_get_interface_addr( netPath->ifaceEndpoint->exanic, netPath->ifaceEndpoint->rx->port_number, &if_addr) != -1)
        memcpy( &frame.ip.srcip, &if_addr.address, sizeof(in_addr_t) );
    memcpy( &frame.ip.dstip, addr, sizeof(in_addr_t) );
    calcIPHdrChecksum( (uint8_t*)&frame.ip, &frame.ip.hdrChecksum, 20 );

    frame.ip.udp.dstPort = htons(port);
    frame.ip.udp.srcPort = htons(port);
    tmp = length+8;
    frame.ip.udp.len = htons(tmp);
    frame.ip.udp.checksum = 0;

    memcpy( &frame.ip.udp.payload, buf, length );
    frameLen = 14 + 20 + 8 + length;

    ts_tx_old = exanic_register_read(netPath->ifaceEndpoint->exanic, REG_PORT_INDEX(netPath->ifaceEndpoint->rx->port_number, REG_PORT_TX_LAST_TIME));
    ts_tx = ts_tx_old;
    ret = endpoint_send(netPath->ifaceEndpoint, (const char*)&frame, frameLen);
    while( ts_tx_old == ts_tx ) // spin on REG_PORT_TX_LAST_TIME waiting for change to indicate packet sent
        ts_tx = exanic_register_read(netPath->ifaceEndpoint->exanic, REG_PORT_INDEX(netPath->ifaceEndpoint->rx->port_number, REG_PORT_TX_LAST_TIME));
    // is this needed any more??
    ts_tx = exanic_register_read(netPath->ifaceEndpoint->exanic, REG_PORT_INDEX(netPath->ifaceEndpoint->rx->port_number, REG_PORT_TX_LAST_TIME));

    // loopback tx frame to rx queue, with timestamp as the time the packet actually left the exanic
    processFrame( &frame, &ts_tx );

    if( ret == 0 )
        return 1;
    return 0;
}

ssize_t netSendEvent( Octet *buf, UInteger16 length, NetPath *netPath, Integer32 alt_dst)
{
    if( netPath->ifaceEndpoint->exanic != NULL )
    {
        if( alt_dst )
            return netSend(buf, length, &alt_dst, netPath, PTP_EVENT_PORT);
        else
            return netSend(buf, length, &netPath->multicastAddr, netPath, PTP_EVENT_PORT);
    }
    else
    {
        ssize_t ret;
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = htons(PTP_EVENT_PORT);

        if (netPath->unicastAddr || alt_dst ) {
            if (netPath->unicastAddr) {
                addr.sin_addr.s_addr = netPath->unicastAddr;
            } else {
                addr.sin_addr.s_addr = alt_dst;
            }

            ret = sendto(netPath->eventSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending uni-cast event message\n");
            /*
             * Need to forcibly loop back the packet since
             * we are not using multicast.
             */
            addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

            ret = sendto(netPath->eventSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error looping back uni-cast event message\n");

        } else {
            addr.sin_addr.s_addr = netPath->multicastAddr;

            ret = sendto(netPath->eventSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending multi-cast event message\n");
        }

        return ret;
    }
}

ssize_t netSendGeneral(Octet *buf, UInteger16 length, NetPath *netPath, Integer32 alt_dst)
{
    if( netPath->ifaceEndpoint->exanic != NULL )
    {
        if( alt_dst )
            return netSend(buf, length, &alt_dst, netPath, PTP_GENERAL_PORT);
        else
            return netSend(buf, length, &netPath->multicastAddr, netPath, PTP_GENERAL_PORT);
    }
    else
    {
        ssize_t ret;
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = htons(PTP_GENERAL_PORT);

        if(netPath->unicastAddr || alt_dst ){
        if (netPath->unicastAddr) {
            addr.sin_addr.s_addr = netPath->unicastAddr;
            } else {
                addr.sin_addr.s_addr = alt_dst;
            }


            ret = sendto(netPath->generalSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending uni-cast general message\n");
        } else {
            addr.sin_addr.s_addr = netPath->multicastAddr;

            ret = sendto(netPath->generalSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending multi-cast general message\n");
        }
        return ret;
    }
}

ssize_t netSendPeerGeneral(Octet *buf, UInteger16 length, NetPath *netPath)
{
    if( netPath->ifaceEndpoint->exanic != NULL )
        return netSend(buf, length, &netPath->peerMulticastAddr, netPath, PTP_GENERAL_PORT);
    else
    {
        ssize_t ret;
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = htons(PTP_GENERAL_PORT);

        if (netPath->unicastAddr) {
            addr.sin_addr.s_addr = netPath->unicastAddr;

            ret = sendto(netPath->generalSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending uni-cast general message\n");

        } else {
            addr.sin_addr.s_addr = netPath->peerMulticastAddr;

            ret = sendto(netPath->generalSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending multi-cast general message\n");
        }
        return ret;
    }
}

ssize_t netSendPeerEvent(Octet *buf, UInteger16 length, NetPath *netPath)
{
    if( netPath->ifaceEndpoint->exanic != NULL )
        return netSend(buf, length, &netPath->peerMulticastAddr, netPath, PTP_EVENT_PORT);
    else
    {
        ssize_t ret;
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = htons(PTP_EVENT_PORT);

        if (netPath->unicastAddr) {
            addr.sin_addr.s_addr = netPath->unicastAddr;

            ret = sendto(netPath->eventSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending uni-cast event message\n");

            /*
             * Need to forcibly loop back the packet since
             * we are not using multicast.
             */
            addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

            ret = sendto(netPath->eventSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error looping back uni-cast event message\n");
        } else {
            addr.sin_addr.s_addr = netPath->peerMulticastAddr;

            ret = sendto(netPath->eventSock, buf, length, 0,
                     (struct sockaddr *)&addr,
                     sizeof(struct sockaddr_in));
            if (ret <= 0)
                DBG("error sending multi-cast event message\n");
        }
        return ret;
    }
}

static void processFrame(struct etherFrameType *frame, uint32_t *timestamp )
{
    int payloadLen;
    struct ptpPacketType *pkt;
    uint8_t *pbuf;
    uint64_t timenanos;

    // Check frame header checksum
    if( checkIPHdrChecksum( (uint8_t*)&frame->ip) != 0 )
        return;

    // Do endian swaps
    frame->ip.udp.dstPort = ntohs(frame->ip.udp.dstPort);
    frame->ip.udp.srcPort = ntohs(frame->ip.udp.srcPort);
    frame->ip.udp.len = ntohs(frame->ip.udp.len);

    // Check if this packet is a multicast
//    if( memcmp( frame->dstmac, multicastMAC, sizeof(frame->dstmac)) == 0 )
    if( memcmp( frame->dstmac, multicastMAC, 3) == 0 )
    {
        if( frame->ip.udp.dstPort == PTP_EVENT_PORT )
        {
            payloadLen = frame->ip.udp.len - 8;
            pkt = (struct ptpPacketType*)malloc(sizeof(struct ptpPacketType));
            pbuf = (uint8_t*)malloc(payloadLen);
            if( (pbuf == NULL) || (pkt == NULL) )
            {
                ERROR("processFrame(): Failed to malloc!\n");
                exit(5);
            }
            pkt->buf = pbuf;
            pkt->len = payloadLen;
            timenanos = exanic_timestamp_to_counter( pLocalNetPath->ifaceEndpoint->exanic, *timestamp );
            pkt->ts_secs = timenanos / (uint64_t)1000000000;
            pkt->ts_nanos = timenanos % (uint64_t)1000000000;
            memcpy(pbuf, &frame->ip.udp.payload, payloadLen );
            if (!netQPut(&pLocalNetPath->eventQ, pkt))
            {
                free(pbuf);
                free(pkt);
                ERROR("netRecvEventCallback: queue full\n");
                return;
            }
        }

        if( frame->ip.udp.dstPort == PTP_GENERAL_PORT )
        {
            payloadLen = frame->ip.udp.len - 8;
            pkt = (struct ptpPacketType*)malloc(sizeof(struct ptpPacketType));
            pbuf = (uint8_t*)malloc(payloadLen);
            if( (pbuf == NULL) || (pkt == NULL) )
            {
                ERROR("processFrame(): Failed to malloc!\n");
                exit(5);
            }
            pkt->buf = pbuf;
            pkt->len = payloadLen;
            timenanos = exanic_timestamp_to_counter( pLocalNetPath->ifaceEndpoint->exanic, *timestamp );
            pkt->ts_secs = timenanos / (uint64_t)1000000000;
            pkt->ts_nanos = timenanos % (uint64_t)1000000000;
            memcpy(pbuf, &frame->ip.udp.payload, payloadLen );
            if (!netQPut(&pLocalNetPath->generalQ, pkt))
            {
                free(pbuf);
                free(pkt);
                ERROR("netRecvEventCallback: queue full\n");
                return;
            }
        }
    }
}

void exanic_Periodic_Handle()
{
    // This function checks the exanic RX buffer to see if we've received any data we're interested in
    int len, iter;
    uint8_t buf[128];
    uint32_t timestamp;
    struct etherFrameType frame;

    memset(&frame, 0, sizeof(frame));
    iter = 100;
    do {
        len = endpoint_recv( pLocalNetPath->ifaceEndpoint, (char*)buf, sizeof(buf), &iter, &timestamp );
        if( len > 0 )
        {
            memcpy( &frame, buf, sizeof(frame) );
            processFrame( &frame, &timestamp );
        }
    } while( len > 0 );
}

/*
 * refresh IGMP on a timeout
 */
/*
 * @return TRUE if successful
 */
Boolean
netRefreshIGMP(NetPath * netPath, RunTimeOpts * rtOpts, PtpClock * ptpClock)
{
    DBG("netRefreshIGMP\n");

    netShutdownMulticast(netPath);

    /* suspend process 100 milliseconds, to make sure the kernel sends the IGMP_leave properly */
    usleep(100*1000);

    if (!netInitMulticast(netPath, rtOpts)) {
        return FALSE;
    }

    INFO("refreshed IGMP multicast memberships\n");
    return TRUE;
}

