#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>

#include <exanic/exanic.h>
#include <exanic/config.h>
#include <exanic/fifo_rx.h>
#include <exanic/time.h>

#include "pcap-structures.h"

int run = 1;

void signal_handler(int signum)
{
    run = 0;
}

/* Parses a string of the format "<device>:<port>" */
int parse_device_port(const char *str, char *device, int *port_number)
{
    char *p, *q;

    p = strchr(str, ':');
    if (p == NULL)
        return -1;

    if ((p-str) >= 16)
        return -1;
    strncpy(device, str, p - str);
    device[p - str] = '\0';
    *port_number = strtol(p + 1, &q, 10);
    if (*(p + 1) == '\0' || *q != '\0')
        /* strtol failed */
        return -1;
    return 0;
}

int main(int argc, char *argv[])
{
    const char *interface = NULL;
    char device[16];
    int port_number;
    exanic_t *exanic;
    exanic_rx_t *rx;
    char rx_buf[16384];
    ssize_t rx_size;
    uint32_t timestamp;
    struct timespec ts;
    unsigned long rx_success = 0, rx_aborted = 0, rx_corrupt = 0, rx_hwovfl = 0, rx_swovfl = 0, rx_other = 0;
    uint64_t *counter_hw;
    uint64_t *counter_sw;
    int64_t threshold = 0;
    unsigned int count = 100000;
    unsigned int data_points = 0;
    unsigned int i;
    int verbose = 0;
    int c;

    while ((c = getopt(argc, argv, "i:n:t:vh?")) != -1)
    {
        switch (c)
        {
            case 'i':
                interface = optarg;
                break;
            case 'n':
                count = atoi(optarg);
                break;
            case 't':
                threshold = atoi(optarg);
                break;
            case 'v':
                verbose = 1;
                break;
            default:
                goto usage_error;
        }
    }

    if (!interface)
        goto usage_error;

    if (exanic_find_port_by_interface_name(interface, device, 16, &port_number) != 0
            && parse_device_port(interface, device, &port_number) != 0)
    {
        fprintf(stderr, "%s: no such interface or not an ExaNIC\n", interface);
        return 1;
    }

    /* Get the exanic handle */
    exanic = exanic_acquire_handle(device);
    if (exanic == NULL)
    {
        fprintf(stderr, "%s: %s\n", device, exanic_get_last_error());
        goto err_acquire_handle;
    }

    /* Get the exanic rx handle */
    rx = exanic_acquire_rx_buffer(exanic, port_number, 0);
    if (rx == NULL)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number,
                exanic_get_last_error());
        goto err_acquire_rx;
    }

    signal(SIGHUP, signal_handler);
    signal(SIGINT, signal_handler);
    signal(SIGPIPE, signal_handler);
    signal(SIGALRM, signal_handler);
    signal(SIGTERM, signal_handler);

    counter_hw = calloc(count, sizeof(uint64_t));
    counter_sw = calloc(count, sizeof(uint64_t));

    /* Start reading from the rx buffer */
    while (run)
    {
        rx_size = exanic_receive_frame(rx, rx_buf, sizeof(rx_buf),
                &timestamp);
        if (rx_size > 0)
        {
            clock_gettime(CLOCK_REALTIME, &ts);
            counter_sw[data_points] = ts.tv_sec * 1000000000 + ts.tv_nsec;
            counter_hw[data_points] = exanic_timestamp_to_counter(exanic, timestamp);
            if (counter_sw[data_points]-counter_hw[data_points] >= threshold)
                data_points++;
            if (data_points == count)
                run = 0;
            rx_success++;
        }
        else if (rx_size < 0)
        {
            switch (-rx_size)
            {
                case EXANIC_RX_FRAME_ABORTED:
                    rx_aborted++;
                    break;
                case EXANIC_RX_FRAME_CORRUPT:
                    rx_corrupt++;
                    break;
                case EXANIC_RX_FRAME_HWOVFL:
                    rx_hwovfl++;
                    break;
                case EXANIC_RX_FRAME_SWOVFL:
                    rx_swovfl++;
                    break;
                case EXANIC_RX_FRAME_TRUNCATED:
                    rx_other++;
                    break;
                default:
                    rx_other++;
            }
        }
    }

    fprintf(stderr, "%s: received=%lu corrupt=%lu aborted=%lu hw_lost=%lu sw_lost=%lu other=%lu\n",
                     argv[0], rx_success, rx_corrupt, rx_aborted, rx_hwovfl, rx_swovfl, rx_other);

    for (i = 0; i < data_points; i++)
    {
        if (verbose)
            printf("%ld %lX %lX\n", counter_sw[i]-counter_hw[i], counter_hw[i], counter_sw[i]);
        else
            printf("%ld\n", counter_sw[i]-counter_hw[i]);
    }

    exanic_release_rx_buffer(rx);
    exanic_release_handle(exanic);
    return 0;

err_acquire_rx:
    exanic_release_handle(exanic);
err_acquire_handle:
    return 1;

usage_error:
    fprintf(stderr, "Usage: %s -i interface\n", argv[0]);
    fprintf(stderr, "  -i: specify Linux interface (e.g. eth0) or ExaNIC port name (e.g. exanic0:0)\n");
    fprintf(stderr, "  -w: dump frames to given file in pcap format (- for stdout)\n");
    fprintf(stderr, "  -H: use hardware timestamps (requires exanic-clock-sync or exanic-ptpd)\n");
    fprintf(stderr, "  -N: write nanosecond-resolution pcap format\n");
    return 1;
}
