#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>
#include <syslog.h>
#include <unistd.h>

#include <exanic/exanic.h>
#include <exanic/pcie_if.h>
#include <exanic/register.h>

#include "clock-sync/common.h"
#include "clock-sync/host.h"
#include "clock-sync/pps.h"
#include "clock-sync/exanic.h"

#define SYSLOG_ID "exanic-clock-sync"
#define EXANIC_MAX 16


static int caught_signal = 0;
static int use_syslog = 0;
static int daemonize = 0;
static char *pidfile = NULL;
static char *prog = NULL;


static void signal_handler(int signum)
{
    caught_signal = signum;
}


void log_printf(int priority, const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    if (use_syslog)
        vsyslog(priority, fmt, ap);
    else
    {
        vprintf(fmt, ap);
        printf("\n");
    }
    va_end(ap);
}


struct exanic_state
{
    enum { SYNC_HOST, SYNC_PPS, SYNC_EXANIC } sync_type;
    enum pps_type pps_type;
    exanic_t *exanic;
    exanic_t *exanic_src;
    struct exanic_pps_sync_state *pps_sync;
    struct exanic_host_sync_state *host_sync;
    struct exanic_exanic_sync_state *exanic_sync;
};


int main(int argc, char *argv[])
{
    struct exanic_state s[EXANIC_MAX];
    int fast_poll = 0;
    int using_pps = 0;
    int ret = 0;
    int n = 0;
    int i;
    int pidfd = -1;

    memset(s, 0, sizeof(s));

    prog = argv[0];

    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] == '-')
        {
            /* Option */
            if (strcmp(argv[i], "--daemon") == 0)
                daemonize = 1;
            else if (strcmp(argv[i], "--syslog") == 0)
                use_syslog = 1;
            else if (strcmp(argv[i], "--pidfile") == 0)
            {
                i++;
                if (i >= argc)
                    goto usage_error;
                pidfile = argv[i];
            }
            else
                goto usage_error;
        }
        else
        {
            char buf[256];
            char *p, *exanic_device, *sync_src;
            strncpy(buf, argv[i], 256);
            buf[255] = '\0';

            /* Parse string of the form exanic_device:sync_src */
            p = strchr(buf, ':');
            if (p == NULL)
                goto usage_error;
            *p = '\0';
            exanic_device = buf;
            sync_src = p + 1;
            if (strlen(exanic_device) == 0 || strlen(sync_src) == 0)
                goto usage_error;

            if (strcmp(sync_src, "host") == 0)
            {
                /* Sync to host clock */
                s[n].sync_type = SYNC_HOST;
                s[n].exanic_src = NULL;
            }
            else if (strcmp(sync_src, "pps") == 0 ||
                    strcmp(sync_src, "pps-differential") == 0)
            {
                /* Sync to a PPS signal from GPS using differential input */
                s[n].sync_type = SYNC_PPS;
                s[n].pps_type = PPS_DIFFERENTIAL;
                s[n].exanic_src = NULL;
                using_pps = 1;
            }
            else if (strcmp(sync_src, "pps-single-ended") == 0)
            {
                /* Sync to a PPS signal from GPS using single-ended input */
                s[n].sync_type = SYNC_PPS;
                s[n].pps_type = PPS_SINGLE_ENDED;
                s[n].exanic_src = NULL;
                using_pps = 1;
            }
            else
            {
                /* Sync to exanic clock */
                s[n].sync_type = SYNC_EXANIC;
                if ((s[n].exanic_src = exanic_acquire_handle(sync_src)) == NULL)
                {
                    fprintf(stderr, "%s: %s: %s\n", prog, sync_src,
                            exanic_get_last_error());
                    ret = 1;
                    goto cleanup;
                }
            }

            if ((s[n].exanic = exanic_acquire_handle(exanic_device)) == NULL)
            {
                fprintf(stderr, "%s: %s: %s\n", prog, exanic_device,
                        exanic_get_last_error());
                ret = 1;
                goto cleanup;
            }
            n++;
        }
    }

    if (n == 0)
        goto usage_error;

    if (pidfile != NULL)
    {
        /* Create and open PID file */
        if ((pidfd = open(pidfile, O_WRONLY | O_CREAT | O_EXCL, 0644)) == -1)
        {
            fprintf(stderr, "%s: %s: %s\n", prog, pidfile, strerror(errno));
            ret = 1;
            pidfile = NULL;
            goto cleanup;
        }
    }

    if (daemonize)
    {
        pid_t pid;

        if ((pid = fork()) != 0)
        {
            /* Parent process */
            if (pid == -1)
            {
                fprintf(stderr, "%s: fork: %s\n", prog, strerror(errno));
                return 1;
            }
            return 0;
        }

        /* Daemon process */
        setsid();
        stdin = freopen("/dev/null", "r", stdin);
        stdout = freopen("/dev/null", "w", stdout);
        stderr = freopen("/dev/null", "w", stderr);
        signal(SIGHUP, SIG_IGN);
        signal(SIGINT, signal_handler);
        signal(SIGPIPE, SIG_IGN);
        signal(SIGALRM, SIG_IGN);
        signal(SIGTERM, signal_handler);
    }
    else
    {
        signal(SIGHUP, signal_handler);
        signal(SIGINT, signal_handler);
        signal(SIGPIPE, signal_handler);
        signal(SIGALRM, signal_handler);
        signal(SIGTERM, signal_handler);
    }

    if (pidfile != NULL)
    {
        /* Write PID to PID file */
        char buf[16];

        sprintf(buf, "%d\n", getpid());
        if (write(pidfd, buf, strlen(buf)) != strlen(buf))
            /* not much we can do; avoid warning for unchecked return value */ ;
        close(pidfd);
    }

    if (use_syslog)
        openlog(SYSLOG_ID, LOG_PID, LOG_DAEMON);

    /* Initialise state */
    for (i = 0; i < n; i++)
    {
        if (s[i].sync_type == SYNC_HOST)
            s[i].host_sync = init_exanic_host_sync(s[i].exanic);
        else if (s[i].sync_type == SYNC_PPS)
            s[i].pps_sync = init_exanic_pps_sync(s[i].exanic, s[i].pps_type);
        else if (s[i].sync_type == SYNC_EXANIC)
            s[i].exanic_sync = init_exanic_exanic_sync(s[i].exanic, s[i].exanic_src);
    }

    /* Run loop */
    while (1)
    {
        if (fast_poll)
            usleep(SHORT_POLL_INTERVAL * 1000000);
        else if (using_pps)
            usleep(PPS_POLL_INTERVAL * 1000000);
        else
            usleep(POLL_INTERVAL * 1000000);
        fast_poll = 0;

        for (i = 0; i < n; i++)
        {
            enum sync_status stat;

            if (s[i].host_sync != NULL)
                stat = poll_exanic_host_sync(s[i].host_sync);
            else if (s[i].pps_sync != NULL)
                stat = poll_exanic_pps_sync(s[i].pps_sync);
            else if (s[i].exanic_sync != NULL)
                stat = poll_exanic_exanic_sync(s[i].exanic_sync);
            else
                continue;

            if (stat == SYNC_FAST_POLL)
                fast_poll = 1;
        }

        if (caught_signal != 0)
        {
            log_printf(LOG_INFO, "Caught signal %d, exiting", caught_signal);
            break;
        }
    }

    /* Shutdown */
    for (i = 0; i < n; i++)
    {
        if (s[i].host_sync != NULL)
            shutdown_exanic_host_sync(s[i].host_sync);

        if (s[i].pps_sync != NULL)
            shutdown_exanic_pps_sync(s[i].pps_sync);

        if (s[i].exanic_sync != NULL)
            shutdown_exanic_exanic_sync(s[i].exanic_sync);
    }

cleanup:
    for (i = 0; i < n; i++)
        exanic_release_handle(s[i].exanic);

    if (use_syslog)
        closelog();

    if (pidfile != NULL)
        unlink(pidfile);

    return ret;

usage_error:
    fprintf(stderr, "Usage: %s [<options>] <exanic>:<sync-source> ...\n\n"
            "<sync-source> is one of:\n"
            "  host\n"
            "    Sync to host clock\n"
            "  pps or pps-differential\n"
            "    Sync using external PPS differential signal\n"
            "  pps-single-ended\n"
            "    Sync using external PPS single-ended signal\n"
            "  an ExaNIC device name\n"
            "    Sync to another ExaNIC on this machine\n",
            argv[0]);
    ret = 1;
    goto cleanup;
}
