#include <linux/ethtool.h>
#include <linux/sockios.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <time.h>

#include <exanic/port.h>
#include <exanic/util.h>
#include <exanic/exanic.h>
#include <exanic/config.h>
#include <exanic/register.h>
#include <exanic/sfp.h>
#include <exanic/firewall.h>
#include <exanic/x4/i2c.h>

int parse_number(const char *str)
{
    char *p;
    int num = strtol(str, &p, 0);
    if (*p != '\0')
        return -1;
    else
        return num;
}

int parse_on_off(const char *str)
{
    if (strcmp(str, "on") == 0)
        return 1;
    else if (strcmp(str, "off") == 0)
        return 0;
    else
        return -1;
}

int parse_device_port(const char *str, char *device, int *port_number)
{
    char *p, *q;

    /* Ignore "/dev/" prefix on device name */
    if (strncmp(str, "/dev/", 5) == 0)
        str += 5;

    p = strchr(str, ':');
    if (p == NULL)
    {
        /* No port number provided */
        if (strlen(str) >= 16)
            return -1;
        strcpy(device, str);
        *port_number = -1;
        return 0;
    }
    else
    {
        /* Format is "<device>:<port>" */
        if ((p-str) >= 16)
            return -1;
        strncpy(device, str, p - str);
        device[p - str] = '\0';
        *port_number = strtol(p + 1, &q, 10);
        if (*(p + 1) == '\0' || *q != '\0')
            /* strtol failed */
            return -1;
        return 0;
    }
}

exanic_t * acquire_handle(const char *device)
{
    exanic_t *exanic;
    exanic = exanic_acquire_handle(device);
    if (exanic == NULL)
    {
        fprintf(stderr, "%s: %s\n", device, exanic_get_last_error());
        exit(1);
    }
    if (exanic_check_supported(exanic) == -1)
    {
        /* Print a message but don't exit */
        fprintf(stderr, "%s not supported: %s\n", device, exanic_get_last_error());
    }
    return exanic;
}

void release_handle(exanic_t *exanic)
{
    if (exanic != NULL)
        exanic_release_handle(exanic);
}

int ethtool_ioctl(int fd, char *ifname, void *data)
{
    struct ifreq ifr;

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ - 1);
    ifr.ifr_data = data;

    return ioctl(fd, SIOCETHTOOL, &ifr);
}

int ethtool_get_priv_flags(int fd, char *ifname, uint32_t *flags)
{
    struct ethtool_value val;
    int ret;

    memset(&val, 0, sizeof(val));
    val.cmd = ETHTOOL_GPFLAGS;
    ret = ethtool_ioctl(fd, ifname, &val);
    if (ret == 0)
        *flags = val.data;

    return ret;
}

int ethtool_set_priv_flags(int fd, char *ifname, uint32_t flags)
{
    struct ethtool_value val;

    memset(&val, 0, sizeof(val));
    val.cmd = ETHTOOL_SPFLAGS;
    val.data = flags;

    return ethtool_ioctl(fd, ifname, &val);
}

int ethtool_get_flag_names(int fd, char *ifname, char flag_names[32][ETH_GSTRING_LEN])
{
    struct ethtool_drvinfo drvinfo;
    struct ethtool_gstrings *strings;
    unsigned len;

    /* Get number of flags from driver info */
    memset(&drvinfo, 0, sizeof(drvinfo));
    drvinfo.cmd = ETHTOOL_GDRVINFO;
    if (ethtool_ioctl(fd, ifname, &drvinfo) == -1)
        return -1;

    len = drvinfo.n_priv_flags;
    if (len > 32)
        len = 32;

    /* Get flag names */
    strings = calloc(1, sizeof(struct ethtool_gstrings) + len * ETH_GSTRING_LEN);
    strings->cmd = ETHTOOL_GSTRINGS;
    strings->string_set = ETH_SS_PRIV_FLAGS;
    strings->len = len;
    if (ethtool_ioctl(fd, ifname, strings) == -1)
    {
        free(strings);
        return -1;
    }

    memset(flag_names, 0, 32 * ETH_GSTRING_LEN);
    memcpy(flag_names, strings->data, len * ETH_GSTRING_LEN);

    return 0;
}

void show_device_info(const char *device, int port_number)
{
    int i, first_port, last_port, port_status;
    const char *str;
    exanic_t *exanic;
    exanic_hardware_id_t hw_type;
    exanic_function_id_t function;
    time_t rev_date;
    int fd;

    fd = socket(AF_INET, SOCK_DGRAM, 0);

    exanic = acquire_handle(device);
    hw_type = exanic_get_hw_type(exanic);
    function = exanic_get_function_id(exanic);
    rev_date = exanic_get_hw_rev_date(exanic);

    printf("Device %s:\n", device);
    str = exanic_hardware_id_str(hw_type);
    printf("  Hardware type: %s\n", (str == NULL) ? "unknown" : str);

    if (hw_type == EXANIC_HW_Z1 || hw_type == EXANIC_HW_Z10 ||
        hw_type == EXANIC_HW_X4 || hw_type == EXANIC_HW_X2 )
    {
        uint32_t id, temp, vccint, vccaux;
        double temp_real=0, vccint_real=0, vccaux_real=0;

        id = exanic_register_read(exanic, REG_HW_INDEX(REG_HW_BOARD_ID));
        printf("  Board ID: 0x%02x\n", id);

        temp = exanic_register_read(exanic, REG_HW_INDEX(REG_HW_TEMPERATURE));
        vccint = exanic_register_read(exanic, REG_HW_INDEX(REG_HW_VCCINT));
        vccaux = exanic_register_read(exanic, REG_HW_INDEX(REG_HW_VCCAUX));

        if (hw_type == EXANIC_HW_Z1 || hw_type == EXANIC_HW_Z10)
        {
            temp_real = temp * (503.975 / 1024.0) - 273.15;
            vccint_real = vccint * 3.0 / 1024.0;
            vccaux_real = vccaux * 3.0 / 1024.0;
        }
        else if (hw_type == EXANIC_HW_X4 || hw_type == EXANIC_HW_X2 )
        {
            temp_real = temp * (503.975 / 4096.0) - 273.15;
            vccint_real = vccint * 3.0 / 4096.0;
            vccaux_real = vccaux * 3.0 / 4096.0;
        }

        printf("  Temperature: %.1f C   VCCint: %.2f V   VCCaux: %.2f V\n",
                temp_real, vccint_real, vccaux_real);
    }

    if (hw_type == EXANIC_HW_X4 || hw_type == EXANIC_HW_X2 )
    {
        uint32_t reg, count, tick_hz;
        double rpm, divisor;

        tick_hz = exanic_register_read(exanic, REG_EXANIC_INDEX(REG_EXANIC_CLK_HZ));
        reg = exanic_register_read(exanic, REG_HW_INDEX(REG_HW_FAN_REV_COUNTER));
        count = reg & 0xFFFFFF;
        divisor = pow(2, (reg >> 24));

        rpm = 60 * 0.5 * count * tick_hz / divisor;

        printf("  Fan speed: %.0f RPM\n", rpm);
    }

    str = exanic_function_id_str(function);
    printf("  Function: %s\n", (str == NULL) ? "unknown" : str);

    {
        char buf[32];
        struct tm *tm;
        char *p;

        tm = gmtime(&rev_date);
        asctime_r(tm, buf);
        if ((p = strchr(buf, '\n')) != NULL)
            *p = '\0';

        printf("  Firmware date: %04d%02d%02d (%s)\n", tm->tm_year + 1900,
                tm->tm_mon + 1, tm->tm_mday, buf);
    }

    if (function == EXANIC_FUNCTION_NIC)
    {
        uint32_t pl_cfg;

        pl_cfg = exanic_get_bridging_config(exanic);
        printf("  Bridging: %s\n", (pl_cfg & EXANIC_FEATURE_BRIDGE) ?
                "on (ports 0 and 1)" : "off");
    }

    if (function == EXANIC_FUNCTION_FIREWALL)
    {
        int fw_capable;
        fw_capable = exanic_get_firewall_capability(exanic);
        printf("  Firewall capability: %s\n", (fw_capable) ? "supported" :
                                                            "unsupported");
    }

    if (function == EXANIC_FUNCTION_DEVKIT && exanic_is_devkit_demo(exanic))
    {
        printf("  **************************************************\n");
        printf("  *** WARNING: THIS CARD HAS EVALUATION FIRMWARE ***\n");
        printf("  *** WHICH WILL CEASE TO FUNCTION AFTER 2 HOURS ***\n");
        printf("  **************************************************\n");
    }
 
    if (port_number == -1)
    {
        first_port = 0;
        last_port = exanic_get_num_ports(exanic) - 1;
    }
    else
        first_port = last_port = port_number;

    for (i = first_port; i <= last_port; i++)
    {
        int rx_usable, tx_usable;
        char ifname[64];

        memset(ifname, 0, sizeof(ifname));

        if (!exanic_port_configurable(exanic, i))
            continue;

        printf("  Port %d:\n", i);

        rx_usable = exanic_port_rx_usable(exanic, i);
        tx_usable = exanic_port_tx_usable(exanic, i);

        if ((function == EXANIC_FUNCTION_NIC
                || function == EXANIC_FUNCTION_DEVKIT)
                    && rx_usable)
        {
            exanic_get_interface_name(exanic, i, ifname, sizeof(ifname));
            if (strlen(ifname) > 0)
                printf("    Interface: %s\n", ifname);
        }

        printf("    Port speed: %u Mbps\n", exanic_get_port_speed(exanic, i));

        port_status = exanic_get_port_status(exanic, i);
        printf("    Port status: %s, %s, %s, %s\n",
                (port_status & EXANIC_PORT_STATUS_ENABLED) ?
                    "enabled" : "disabled",
                (port_status & EXANIC_PORT_STATUS_SFP) ?
                    "SFP present" : "no SFP",
                (port_status & EXANIC_PORT_STATUS_SIGNAL) ?
                    "signal detected" : "no signal",
                (port_status & EXANIC_PORT_STATUS_LINK) ?
                    "link active" : "no link");

        if (function == EXANIC_FUNCTION_NIC && exanic_get_num_ports(exanic) == 4 && i < 3 && (rx_usable || tx_usable))
        {
            uint32_t pl_cfg;
            uint32_t rx_bit = 0, tx_bit = 0;

            pl_cfg = exanic_get_bridging_config(exanic);

            switch (i)
            {
                case 0:
                    rx_bit = EXANIC_FEATURE_MIRROR_RX_0;
                    tx_bit = EXANIC_FEATURE_MIRROR_TX_0;
                    break;
                case 1:
                    rx_bit = EXANIC_FEATURE_MIRROR_RX_1;
                    tx_bit = EXANIC_FEATURE_MIRROR_TX_1;
                    break;
                case 2:
                    rx_bit = EXANIC_FEATURE_MIRROR_RX_2;
                    tx_bit = EXANIC_FEATURE_MIRROR_TX_2;
                    break;
            }

            printf("    Mirroring: %s\n",
                    (pl_cfg & rx_bit) && (pl_cfg & tx_bit) ? "RX and TX" :
                    (pl_cfg & rx_bit) ? "RX only" :
                    (pl_cfg & tx_bit) ? "TX only" : "off");
        }

        if ((function == EXANIC_FUNCTION_NIC 
                || function == EXANIC_FUNCTION_DEVKIT)
                    && rx_usable)
        {
            if (hw_type == EXANIC_HW_X4 || hw_type == EXANIC_HW_X2)
            {
                int mac_rules = exanic_register_read(exanic, 
                                    REG_EXTENDED_PORT_INDEX(i, 
                                      REG_EXTENDED_PORT_NUM_MAC_FILTER_RULES));
                int ip_rules = exanic_register_read(exanic,
                                    REG_EXTENDED_PORT_INDEX(i, 
                                      REG_EXTENDED_PORT_NUM_IP_FILTER_RULES));
                printf("    MAC filters: %d", mac_rules);
                printf("  IP filters: %d\n", ip_rules);
            }
            printf("    Promiscuous mode: %s\n",
                    exanic_get_promiscuous_mode(exanic, i) ? "on" : "off");
        }


        if ((function == EXANIC_FUNCTION_NIC 
                || function == EXANIC_FUNCTION_DEVKIT)
                && strlen(ifname) > 0)
        {
            char flag_names[32][ETH_GSTRING_LEN];
            uint32_t flags;
            int b;

            if (ethtool_get_flag_names(fd, ifname, flag_names) == 0 &&
                ethtool_get_priv_flags(fd, ifname, &flags) == 0)
            {
                for (b = 0; b < 32; b++)
                {
                    if (strcmp("bypass_only", flag_names[b]) == 0)
                        printf("    Bypass-only mode: %s\n",
                                (flags & (1 << b)) ? "on" : "off");
                }
            }
        }

        if ((function == EXANIC_FUNCTION_NIC 
                || function == EXANIC_FUNCTION_DEVKIT)
                    && (rx_usable || tx_usable))
        {
            uint8_t mac_addr[6];
            exanic_if_addr_t ifaddr;
            struct in_addr address, netmask;

            memset(mac_addr, 0, sizeof(mac_addr));
            if (exanic_get_mac_addr(exanic, i, mac_addr) == 0)
            {
                printf("    MAC address: %02x:%02x:%02x:%02x:%02x:%02x\n",
                        mac_addr[0], mac_addr[1], mac_addr[2],
                        mac_addr[3], mac_addr[4], mac_addr[5]);
            }
            else
            {
                fprintf(stderr, "%s:%d: error reading MAC address: %s\n",
                        device, i, exanic_get_last_error());
            }

            if (exanic_get_interface_addr(exanic, i, &ifaddr) == 0)
            {
                address.s_addr = ifaddr.address;
                netmask.s_addr = ifaddr.netmask;
                if (address.s_addr != INADDR_ANY)
                {
                    printf("    IP address: %s", inet_ntoa(address));
                    printf("  Mask: %s\n", inet_ntoa(netmask));
                }
            }
        }

        if ((function == EXANIC_FUNCTION_NIC
                || function == EXANIC_FUNCTION_DEVKIT)
                    && (rx_usable || tx_usable))
        {
            exanic_port_stats_t port_stats;

            memset(&port_stats, 0, sizeof(port_stats));
            exanic_get_port_stats(exanic, i, &port_stats);
            printf("    RX packets: %u  ignored: %u  error: %u  dropped: %u\n",
                    port_stats.rx_count, port_stats.rx_ignored_count,
                    port_stats.rx_error_count, port_stats.rx_dropped_count);
            printf("    TX packets: %u\n", port_stats.tx_count);
        }
    }

    release_handle(exanic);
}

void show_all_devices(void)
{
    char device_file[32];
    int device_number;
    for (device_number = 0; ; device_number++)
    {
        snprintf(device_file, sizeof(device_file), "/dev/exanic%u", device_number);
        if (access(device_file, F_OK) != 0)
            break;

        show_device_info(device_file+5, -1);
    }
}

void show_sfp_status(const char *device, int port_number)
{
    int port_status;
    exanic_t *exanic;
    exanic_sfp_info_t info;
    exanic_sfp_diag_info_t diag_info;

    exanic = acquire_handle(device);
    port_status = exanic_get_port_status(exanic, port_number);

    if ((port_status & EXANIC_PORT_STATUS_SFP) == 0)
    {
        fprintf(stderr, "%s:%d: SFP not present\n", device, port_number);
        release_handle(exanic);
        exit(1);
    }

    printf("Device %s port %d SFP status:\n", device, port_number);

    if (exanic_get_sfp_info(exanic, port_number, &info) == 0)
    {
        printf("  Vendor: %16.16s PN: %16.16s  rev: %4.4s\n",
               info.vendor_name, info.vendor_pn, info.vendor_rev);
        printf("                           SN: %16.16s date: %8.8s\n",
                info.vendor_sn, info.date_code);

        printf("  Wavelength: %d nm\n", info.wavelength);
        printf("  Nominal bit rate: %d Mbps\n", info.bit_rate);
    }
    else
    {
        printf("  SFP EEPROM not available\n");
    }

    if (exanic_get_sfp_diag_info(exanic, port_number, &diag_info) == 0)
    {
        printf("  Rx power: %.1f dBm (%.2f mW)\n",
                log10(diag_info.rx_power) * 10, diag_info.rx_power);
        printf("  Tx power: %.1f dBm (%.2f mW)\n",
                log10(diag_info.tx_power) * 10, diag_info.tx_power);
        printf("  Temperature: %.1f C\n", diag_info.temp);
    }
    else
    {
        printf("  SFP diagnostics not available\n");
    }

    release_handle(exanic);
}

void get_interface_name(const char *device, int port_number,
                        char *buf, size_t len)
{
    exanic_t *exanic = acquire_handle(device);
    if (exanic_get_interface_name(exanic, port_number, buf, len) != 0)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number,
                exanic_get_last_error());
        release_handle(exanic);
        exit(1);
    }
    release_handle(exanic);
}

void set_port_enable_state(const char *device, int port_number, int mode)
{
    struct ifreq ifr;
    int fd;

    get_interface_name(device, port_number, ifr.ifr_name, IFNAMSIZ);

    /* Enable port via socket ioctls */
    if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ||
        ioctl(fd, SIOCGIFFLAGS, &ifr) == -1)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number, strerror(errno));
        exit(1);
    }

    if (mode)
        ifr.ifr_flags |= IFF_UP;
    else
        ifr.ifr_flags &= ~IFF_UP;

    if (ioctl(fd, SIOCSIFFLAGS, &ifr) == -1)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number, strerror(errno));
        exit(1);
    }

    printf("%s:%d: port %s\n", device, port_number,
            mode ? "enabled" : "disabled");
}

void set_ethtool_priv_flags(const char *device, int port_number,
                            const char *flag_name, int mode)
{
    char ifname[IFNAMSIZ];
    char flag_names[32][ETH_GSTRING_LEN];
    uint32_t flags = 0;
    unsigned i;
    int fd;

    get_interface_name(device, port_number, ifname, IFNAMSIZ);

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd == -1)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number, strerror(errno));
        exit(1);
    }

    /* Get flag names and current setting */
    if (ethtool_get_flag_names(fd, ifname, flag_names) == -1 ||
        ethtool_get_priv_flags(fd, ifname, &flags) == -1)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number, strerror(errno));
        exit(1);
    }

    /* Look for flag name */
    for (i = 0; i < 32; i++)
        if (strcmp(flag_name, flag_names[i]) == 0)
            break;
    if (i == 32)
    {
        fprintf(stderr, "%s:%d: could not find flag %s\n", device, port_number,
                flag_name);
        exit(1);
    }

    if (mode)
        flags |= (1 << i);
    else
        flags &= ~(1 << i);

    /* Set flags */
    if (ethtool_set_priv_flags(fd, ifname, flags) == -1)
    {
        fprintf(stderr, "%s:%d: %s\n", device, port_number, strerror(errno));
        exit(1);
    }

    close(fd);
}

void set_firewall_state(const char *device, exanic_firewall_state_t state)
{
    exanic_t *exanic = acquire_handle(device);
    if (exanic_set_firewall_state(exanic, state) != 0)
    {
        fprintf(stderr, "%s: error changing firewall state: %s\n",
                device, exanic_get_last_error());
        release_handle(exanic);
        exit(1);
    }
    if (state == EXANIC_FIREWALL_ENABLE)
        printf("%s: firewall enabled\n", device);
    else if (state == EXANIC_FIREWALL_DISABLE)
        printf("%s: firewall disabled\n", device);
    else if (state == EXANIC_FIREWALL_TRANSPARENT)
        printf("%s: firewall in transparent mode\n", device);
    release_handle(exanic);
}

void set_phy_local_loopback(const char * device, int port_number, int enable)
{
    exanic_t *exanic = acquire_handle(device);
    char buf;
    uint8_t reg_addr;

    reg_addr = 0x0A;

    if (exanic_x4_i2c_phy_read(exanic, port_number, reg_addr, &buf, 1) != 0)
    {
        fprintf(stderr, "%s:%d: error reading from PHY\n", device, port_number);
        goto out;
    }

    if (enable)
        buf = buf & (~0x40);
    else
        buf = buf | 0x40;

    if (exanic_x4_i2c_phy_write(exanic, port_number, reg_addr, &buf, 1) != 0)
    {
        fprintf(stderr, "%s:%d: error writing to PHY\n", device, port_number);
        goto out;
    }

out:
    release_handle(exanic);
}

static const struct phy_parameter {
    const char *name;
    uint8_t reg_addr;
    uint8_t mask;
} phy_parameters[] = {
    { "rx-gain",        0x10, 0x7f },
    { "rx-preemphasis", 0x11, 0x1f },
    { "rx-offset",      0x12, 0xff },
    { "tx-gain",        0x16, 0x07 },
    { "tx-preemphasis", 0x17, 0x1f },
    { "tx-slewrate",    0x18, 0x07 }
};

void set_phy_parameter(const char *device, int port_number,
                       const char *parameter_name, const char *value_string)
{
    exanic_t *exanic = acquire_handle(device);
    uint8_t reg_addr = 0;
    uint8_t mask = 0;
    char buf;
    int i;

    for (i = 0; i < sizeof(phy_parameters)/sizeof(struct phy_parameter); i++)
    {
        if (strcmp(parameter_name, phy_parameters[i].name) == 0)
        {
            reg_addr = phy_parameters[i].reg_addr;
            mask = phy_parameters[i].mask;
        }
    }

    if (!mask)
    {
        fprintf(stderr, "%s:%d: invalid parameter name %s\n", device,
                port_number, parameter_name);
        goto out;
    }

    if (value_string)
    {
        int value = parse_number(value_string);
        if ((value < 0) || (value > mask))
        {
            fprintf(stderr, "%s:%d: invalid value specified for %s\n", device,
                    port_number, parameter_name);
            goto out;
        }

        buf = value;
        if (exanic_x4_i2c_phy_write(exanic, port_number, reg_addr, &buf, 1) != 0)
        {
            fprintf(stderr, "%s:%d: error writing to PHY\n", device, port_number);
            goto out;
        }
    }
    else
    {
        if (exanic_x4_i2c_phy_read(exanic, port_number, reg_addr, &buf, 1) != 0)
        {
            fprintf(stderr, "%s:%d: error reading from PHY\n", device, port_number);
            goto out;
        }
    }
    printf("%s:%d: %s = %u (range 0..%u)\n", device, port_number,
                                                  parameter_name, buf, mask);
out:
    release_handle(exanic);
}

void show_phy_parameters(const char *device, int port_number)
{
    exanic_t *exanic = acquire_handle(device);
    const char *parameter_name;
    uint8_t reg_addr;
    uint8_t mask;
    char buf;
    int i;

    for (i = 0; i < sizeof(phy_parameters)/sizeof(struct phy_parameter); i++)
    {
        parameter_name = phy_parameters[i].name;
        reg_addr = phy_parameters[i].reg_addr;
        mask = phy_parameters[i].mask;
        if (exanic_x4_i2c_phy_read(exanic, port_number, reg_addr, &buf, 1) != 0)
        {
            fprintf(stderr, "%s:%d: error reading from PHY\n", device, port_number);
            break;
        }
 
        printf("%s:%d: %s = %u (range 0..%u)\n", device, port_number,
                                                  parameter_name, buf, mask);
    }
    release_handle(exanic);
}
 
void show_firewall_filters(const char *device)
{
    int i;
    int num_filters, max_filters;
    exanic_t *exanic;
    char filter[256];

    exanic = acquire_handle(device);
    num_filters = 0;
    max_filters = exanic_get_num_firewall_filters(exanic);
    if (max_filters == -1)
    {
        fprintf(stderr, "%s: %s\n", device, exanic_get_last_error());
        release_handle(exanic);
        exit(1);
    }

    printf("Firewall filters on %s:\n", device);

    for (i = 0; i < max_filters; i++)
    {
        if (exanic_get_firewall_filter(exanic, i, filter, sizeof(filter)) == -1)
        {
            fprintf(stderr, "%s: error on filter %d: %s\n",
                    device, i, exanic_get_last_error());
            release_handle(exanic);
            exit(1);
        }
        if (strlen(filter) > 0)
        {
            printf("%4u: %s\n", i, filter);
            num_filters++;
        }
    }

    printf("%d used, %d total\n", num_filters, max_filters);

    release_handle(exanic);
}


void show_firewall_dump(const char *device)
{
    int i;
    int max_filters;
    exanic_t *exanic;
    char filter[256];
    exanic_firewall_state_t fw_state;

    exanic = acquire_handle(device);
    fw_state = exanic_get_firewall_state(exanic);
    max_filters = exanic_get_num_firewall_filters(exanic);
    if (max_filters == -1)
    {
        fprintf(stderr, "%s: %s\n", device, exanic_get_last_error());
        release_handle(exanic);
        exit(1);
    }

    printf("filters:%d\n", max_filters);

    if (fw_state == EXANIC_FIREWALL_DISABLE)
        printf("mode:disabled\n");
    else if (fw_state == EXANIC_FIREWALL_ENABLE)
        printf("mode:enabled\n");
    else if (fw_state == EXANIC_FIREWALL_TRANSPARENT)
        printf("mode:transparent\n");

    for (i = 0; i < max_filters; i++)
    {
        if (exanic_get_firewall_filter(exanic, i, filter, sizeof(filter)) == -1)
        {
            fprintf(stderr, "%s: error on filter %d: %s\n",
                    device, i, exanic_get_last_error());
            release_handle(exanic);
            exit(1);
        }
        if (strlen(filter) > 0)
            printf("%u:%s\n", i, filter);
    }

    release_handle(exanic);
}

void set_firewall_filter(const char *device, int slot, const char *filter)
{
    char buf[256];
    exanic_t *exanic = acquire_handle(device);
    if (exanic_get_firewall_filter(exanic, slot, buf, sizeof(buf)) == 0
            && strlen(buf) > 0)
    {
        fprintf(stderr, "%s: error setting firewall filter: "
                "filter %d is already in use\n",
                device, slot);
        release_handle(exanic);
        exit(1);
    }
    if (exanic_set_firewall_filter(exanic, slot, filter) == -1)
    {
        fprintf(stderr, "%s: error setting firewall filter: %s\n",
                device, exanic_get_last_error());
        release_handle(exanic);
        exit(1);
    }
    printf("%s: added firewall filter %d: %s\n", device, slot, filter);
    release_handle(exanic);
}

void clear_firewall_filter(const char *device, int slot)
{
    exanic_t *exanic = acquire_handle(device);
    if (exanic_clear_firewall_filter(exanic, slot) == -1)
    {
        fprintf(stderr, "%s: error deleting firewall filter: %s\n",
                device, exanic_get_last_error());
        release_handle(exanic);
        exit(1);
    }
    printf("%s: deleted firewall filter %d\n", device, slot);
    release_handle(exanic);
}

void clear_all_firewall_filters(const char *device)
{
    exanic_t *exanic = acquire_handle(device);
    if (exanic_clear_all_firewall_filters(exanic) == -1)
    {
        fprintf(stderr, "%s: error deleting firewall filters: %s\n",
                device, exanic_get_last_error());
        release_handle(exanic);
        exit(1);
    }
    printf("%s: deleted all firewall filters\n", device);
    release_handle(exanic);
}


int main(int argc, char *argv[])
{
    char device[16];
    int port_number;
    int mode;

    if (argc < 2)
    {
        show_all_devices();
        return 0;
    }

    if (argv[1][0] == '-')
        goto usage_error;

    if (exanic_find_port_by_interface_name(argv[1], device, 16, &port_number)
            != 0 && parse_device_port(argv[1], device, &port_number) != 0)
        goto usage_error;

    if (argc == 2)
    {
        show_device_info(device, port_number);
        return 0;
    }
    else if (argc == 4 && strcmp(argv[2], "sfp") == 0
            && strcmp(argv[3], "status") == 0 && port_number != -1)
    {
        show_sfp_status(device, port_number);
        return 0;
    }
    else if (argc == 3 && strcmp(argv[2], "up") == 0 && port_number != -1)
    {
        set_port_enable_state(device, port_number, 1);
        return 0;
    }
    else if (argc == 3 && strcmp(argv[2], "down") == 0 && port_number != -1)
    {
        set_port_enable_state(device, port_number, 0);
        return 0;
    }
    else if (argc == 4 && strcmp(argv[2], "bridging") == 0 && port_number == -1)
    {
        if ((mode = parse_on_off(argv[3])) == -1)
            goto usage_error;
        set_ethtool_priv_flags(device, 0, "bridging", mode);
        printf("%s: bridging %s\n", device,
                mode ? "on (ports 0 and 1)" : "off");
        return 0;
    }
    else if (argc == 4 && strcmp(argv[2], "mirror-rx") == 0 &&
            port_number != -1)
    {
        if ((mode = parse_on_off(argv[3])) == -1)
            goto usage_error;
        set_ethtool_priv_flags(device, port_number, "mirror_rx", mode);
        printf("%s:%d: RX mirroring %s\n", device, port_number,
                mode ? "on" : "off");
        return 0;
    }
    else if (argc == 4 && strcmp(argv[2], "mirror-tx") == 0 &&
            port_number != -1)
    {
        if ((mode = parse_on_off(argv[3])) == -1)
            goto usage_error;
        set_ethtool_priv_flags(device, port_number, "mirror_tx", mode);
        printf("%s:%d: TX mirroring %s\n", device, port_number,
                mode ? "on" : "off");
        return 0;
    }
    else if (argc == 4 && strcmp(argv[2], "bypass-only") == 0 &&
            port_number != -1)
    {
        if ((mode = parse_on_off(argv[3])) == -1)
            goto usage_error;
        set_ethtool_priv_flags(device, port_number, "bypass_only", mode);
        printf("%s:%d: bypass-only mode %s, kernel RX and TX %s\n",
                device, port_number, mode ? "on" : "off",
                mode ? "disabled" : "enabled");
        return 0;
    }
    else if ((argc == 3 || argc == 4) &&
             (strncmp(argv[2], "rx-", 3) == 0 ||
              strncmp(argv[2], "tx-", 3) == 0) && port_number != -1)
    {
        set_phy_parameter(device, port_number, argv[2], (argc == 4) ? argv[3] : NULL);
        return 0;
    }
    else if (argc == 3 && strcmp(argv[2], "show-phy-param") == 0 &&
                port_number != -1)
    {
        show_phy_parameters(device, port_number);
        return 0;
    }
    else if (argc == 4 && strcmp(argv[2], "local-loopback") == 0 &&
                port_number != -1)
    {
        if ((mode = parse_on_off(argv[3])) == -1)
            goto usage_error;
        set_phy_local_loopback(device, port_number, mode);
        printf("%s:%d: local-loopback mode %s\n", device, port_number, 
                mode ? "enabled" : "disabled");
        return 0;
    }
    /* below commands for firewall firmware only */
    else if (argc == 4 && strcmp(argv[2], "firewall") == 0
            && port_number == -1)
    {
        if (strcmp(argv[3], "on") == 0)
            set_firewall_state(device, EXANIC_FIREWALL_ENABLE);
        else if (strcmp(argv[3], "off") == 0)
            set_firewall_state(device, EXANIC_FIREWALL_DISABLE);
        else if (strcmp(argv[3], "transparent") == 0)
            set_firewall_state(device, EXANIC_FIREWALL_TRANSPARENT);
        else
            goto usage_error;
        return 0;
    }
    else if (argc == 6 && strcmp(argv[2], "filter") == 0
            && strcmp(argv[3], "add") == 0 && port_number == -1)
    {
        int slot = parse_number(argv[4]);
        if (slot == -1)
            goto usage_error;
        set_firewall_filter(device, slot, argv[5]);
        return 0;
    }
    else if (argc == 5 && strcmp(argv[2], "filter") == 0
            && strcmp(argv[3], "del") == 0 && port_number == -1)
    {
        if (strcmp(argv[4], "all") == 0)
            clear_all_firewall_filters(device);
        else
        {
            int slot = parse_number(argv[4]);
            if (slot == -1)
                goto usage_error;
            clear_firewall_filter(device, slot);
        }
        return 0;
    }
    else if (argc == 4 && strcmp(argv[2], "filter") == 0
            && strcmp(argv[3], "list") == 0 && port_number == -1)
    {
        show_firewall_filters(device);
        return 0;
    }
    else if (argc == 3 && strcmp(argv[2], "firewall-dump") == 0
            && port_number == -1)
    {
        show_firewall_dump(device);
        return 0;
    }

usage_error:
    fprintf(stderr, "exanic-config version 1.6.0\n");
    fprintf(stderr, "Detailed network interface configuration and status:\n");
    fprintf(stderr, "   %s <device>\n", argv[0]);
    fprintf(stderr, "   %s <device> sfp status\n", argv[0]);
    fprintf(stderr, "   %s <device> { up | down }\n", argv[0]);
    fprintf(stderr, "   %s <device> bridging { on | off }\n", argv[0]);
    fprintf(stderr, "   %s <device> mirror-rx { on | off }\n", argv[0]);
    fprintf(stderr, "   %s <device> mirror-tx { on | off }\n", argv[0]);
    fprintf(stderr, "   %s <device> local-loopback { on | off }\n", argv[0]);
    fprintf(stderr, "   %s <device> bypass-only { on | off }\n", argv[0]);
    fprintf(stderr, "      <device> can be a Linux interface name, ExaNIC device (e.g. exanic0)\n");
    fprintf(stderr, "      or ExaNIC device:port (e.g. exanic0:0)\n");
    return 1;
}
